#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#include "constants.h"

/* Light a led based on the chosen player and the positions. */
void lightLED(char player, int line, int col) {
	if (player == NO_PLAYER)
		return;
	PORTA = 0;
	PORTC = 0;
	PORTB = 0b11111111;

	if (game_matrix[line][col] == NO_PLAYER) {
		if (player == PLAYER_RED) {
			PORTA = (1 << line_pins_red[line]);
			PORTB &= ~(1 << col_ports[col]);

		} else if (player == PLAYER_BLUE) {
			PORTC = (1 << line_pins_blue[line]);
			PORTB &= ~(1 << col_ports[col]);
		}
	} else {
		PORTA = (1 << line_pins_red[line]);
		PORTC = (1 << line_pins_blue[line]);
		PORTB &= ~(1 << col_ports[col]);
	}
}

/* Light a line based on the chosen player. */
void lightLine(char player, char line[MATRIX_LEN], char line_number) {
	if (player == NO_PLAYER)
		return;

	char positions = 0b11111111;
	int i;

	PORTA = 0;
	PORTC = 0;

	for (i = 0; i < MATRIX_LEN; ++i) {
		if (line[i] == player) {
			positions &= ~(1 << (MATRIX_LEN - 1 - i));
		}
	}

	if (player == PLAYER_RED)
		PORTA = (1 << (MATRIX_LEN - 1 - line_number));
	else
		PORTC = (1 << (MATRIX_LEN - 1 - line_number));
	PORTB = positions;
}

/*
 * Initialise the rows and collumns of the LED matrix
 * and the buttons.
 */
void IO_init() {
	int i, j;
	for (i = 0; i < MATRIX_LEN; ++i) {
		DDRA |= (1 << line_pins_red[i]);
		DDRB |= (1 << col_ports[i]);
		DDRC |= (1 << line_pins_blue[i]);
	}

	PORTD = 0;
	for (i = 0; i < NR_BUTTONS; ++i) {
		DDRD &= ~(1 << button_ports[i]);
		PORTD |= (1 << button_ports[i]);
	}

	for (i = 0; i < MATRIX_LEN; ++i) {
		for (j = 0; j < MATRIX_LEN; ++j) {
			game_matrix[i][j] = NO_PLAYER;
		}
	}

	for (i = 0; i < MATRIX_LEN * MATRIX_LEN; ++i)
		visited[i] = UNVISITED;
}

/* Timer at 1 kHZ with prescaler 1 */
void timers_init() {
	sei();

	OCR1A = 0x2710;
	OCR0A = 0xFA;

	TIMSK1 = (1 << OCIE1A);
	TIMSK0 = (1 << OCIE0A);

	TCCR0B = (1 << CS01) | (1 << CS00) | (1 << WGM02);
	TCCR1B = (1 << WGM12) | (1 << CS10);
}

int compute_points(char player) {
	int sum = 0;
	if (player == PLAYER_RED) {
		sum += red_captures;
		for (int i = 0; i < MATRIX_LEN; ++i) {
			for (int j = 0; j < MATRIX_LEN; ++j) {
				if (game_matrix[i][j] == PLAYER_RED)
					sum += 1;
			}
		}
	} else if (player == PLAYER_BLUE) {
		sum += blue_captures;
		for (int i = 0; i < MATRIX_LEN; ++i) {
			for (int j = 0; j < MATRIX_LEN; ++j) {
				if (game_matrix[i][j] == PLAYER_BLUE)
					sum += 1;
			}
		}
	}
	return sum;
}

/*
 * Change the position only at the release of the button.
 */
void navigate() {
	if ((PIND & (1 << button_ports[DOWN_BUTTON])) == 0) {
		button_state = 0;
		button_type = DOWN_BUTTON;
	} else if ((PIND & (1 << button_ports[UP_BUTTON])) == 0) {
		button_state = 0;
		button_type = UP_BUTTON;
	} else if ((PIND & (1 << button_ports[LEFT_BUTTON])) == 0) {
		button_state = 0;
		button_type = LEFT_BUTTON;
	} else if ((PIND & (1 << button_ports[RIGHT_BUTTON])) == 0) {
		button_state = 0;
		button_type = RIGHT_BUTTON;
	} else if ((PIND & (1 << button_ports[SELECT_BUTTON])) == 0) {
		counter++;
		if (counter == PASS_COUNTER) {
			if (turn == PLAYER_RED) {
				player_red_pass = 0;
				turn = PLAYER_BLUE;
				currentLine = currentCol = 0;
				counter = 0;
			}
			else if (turn == PLAYER_BLUE) {
				player_blue_pass = 0;
				turn = PLAYER_RED;
				currentLine = currentCol = 0;
				counter = 0;
			}
			if (player_red_pass == 0 && player_blue_pass == 0) {
				int red_points = compute_points(PLAYER_RED);
				int blue_points = compute_points(PLAYER_BLUE);
				winner = (red_points > blue_points ? PLAYER_RED : PLAYER_BLUE);
			}
		}
		button_state = 0;
		button_type = SELECT_BUTTON;
	} else if (!button_state) {
		button_state = 1;
		counter = 0;
		player_red_pass = -1;
		player_blue_pass = -1;
	} else {
		button_state = -1;
		button_type = -1;
	}

	if (button_state == 1) {
		switch (button_type) {
			case DOWN_BUTTON:
				currentLine++;
					if (currentLine >= MATRIX_LEN)
						currentLine = 0;
				break;
			case UP_BUTTON:
				currentLine--;
				if (currentLine < 0)
					currentLine = MATRIX_LEN - 1;
				break;
			case LEFT_BUTTON:
				currentCol--;
				if (currentCol < 0)
					currentCol = MATRIX_LEN - 1;
				break;
			case RIGHT_BUTTON:
				currentCol++;
				if (currentCol >= MATRIX_LEN)
					currentCol = 0;
				break;
			case SELECT_BUTTON:
				if (game_matrix[currentLine][currentCol] != NO_PLAYER)
					return;
				game_matrix[currentLine][currentCol] = turn;
				if (turn == PLAYER_RED)
					turn = PLAYER_BLUE;
				else
					turn = PLAYER_RED;
				currentLine = currentCol = 0;
				break;
		}
	}
}

int check_neighbour(char oponent, int line, int col) {
	char player = turn;
	char left_neighb;
	char right_neighb;
	char up_neighb;
	char down_neighb;
	int ret_l = -1, ret_r = -1, ret_u = -1, ret_d = -1;

	if (visited[line * MATRIX_LEN + col] == VISITED)
		return 0;

	visited[line * MATRIX_LEN + col] = VISITED;
	left_neighb = (col > 0 ? game_matrix[line][col - 1] : -1);
	right_neighb = (col < MATRIX_LEN - 1 ? game_matrix[line][col + 1] : -1);
	up_neighb = (line > 0 ? game_matrix[line - 1][col] : -1);
	down_neighb = (line < MATRIX_LEN - 1 ? game_matrix[line + 1][col] : -1);

	if (left_neighb == NO_PLAYER
		|| right_neighb == NO_PLAYER
		|| up_neighb == NO_PLAYER
		|| down_neighb == NO_PLAYER)
		return -1;

	if ((left_neighb == oponent || left_neighb == -1)
		&& (right_neighb == oponent || right_neighb == -1)
		&& (up_neighb == oponent || up_neighb == -1)
		&& (down_neighb == oponent || down_neighb == -1)) {
		if (oponent == PLAYER_RED)
			red_captures += 1;
		else if (oponent == PLAYER_BLUE)
			blue_captures += 1;
		game_matrix[line][col] = NO_PLAYER;
		return 1;
	}

	if (left_neighb == player) {
		ret_l = check_neighbour(oponent, line, col - 1);
		if (ret_l == -1)
			return ret_l;
	}

	if (right_neighb == player) {
		ret_r = check_neighbour(oponent, line, col + 1);
		if (ret_r == -1)
			return ret_r;
	}

	if (up_neighb == player) {
		ret_u = check_neighbour(oponent, line - 1, col);
		if (ret_u == -1)
			return ret_u;
	}

	if (down_neighb == player) {
		ret_d = check_neighbour(oponent, line + 1, col);
		if (ret_d == -1)
			return ret_d;
	}

	if (oponent == PLAYER_RED)
		red_captures += 1;
	else if (oponent == PLAYER_BLUE)
		blue_captures += 1;
	game_matrix[line][col] = NO_PLAYER;
	return 0;
}

/*
 * At each interrupt of the timer light up a row of LEDs (one interrupt
 * for the red ones and the other for the blue).
 */
ISR(TIMER1_COMPA_vect) {
	if (winner != -1) {
		turn = NO_PLAYER;
		for (int i = 0; i < MATRIX_LEN; ++i) {
			for (int j = 0; j < MATRIX_LEN; ++j) {
				if (winner == PLAYER_RED)
					game_matrix[i][j] = red_winner[i][j];
				else
					game_matrix[i][j] = blue_winner[i][j];
			}
		}
	}
	if (lineCounter >= MATRIX_LEN) {
		lightLED(turn, currentLine, currentCol);
		lineCounter = 0;
	} else {
		lightLine(light_flag, game_matrix[lineCounter], lineCounter);
		if (light_flag == PLAYER_RED) {
			light_flag = PLAYER_BLUE;
		} else if (light_flag == PLAYER_BLUE) {
			light_flag = PLAYER_RED;
			lineCounter++;
		}
	}
}

/*
 * Timer 0 interrupt for button checking.
 */
ISR(TIMER0_COMPA_vect) {
	navigate();
	if (button_type == SELECT_BUTTON) {
		char player = (turn == PLAYER_RED ? PLAYER_BLUE : PLAYER_RED);
		for (int i = 0; i < MATRIX_LEN; ++i) {
			for (int j = 0; j < MATRIX_LEN; ++j) {
				if (game_matrix[i][j] == turn)
					check_neighbour(player, i, j);
			}
		}
		for (int i = 0; i < MATRIX_LEN * MATRIX_LEN; ++i)
			visited[i] = UNVISITED;
	}
}

int main(void) {
	IO_init();

	timers_init();

	while (1);

	return 0;
}
